// Split all sections into different methods, ready to call
// Defined axis: [roll,pitch]

// Uncomment to debug
//#define DEBUG 1
//#define POSITIONFEEDBACK 1
//#define DIRECTIONFEEDBACK 1
//#define STEPTIMEFEEDBACK
//#define USERCONTROL 1
//#define LOOPUPDATE 1

// Test type:
int testType = 0;
// 0 = Simtools 16bit mode (no feedback)
// 1 = Speed Test Mode (terminal interface)
// 2 = ???
// 3 = ???

// Declare pins
int EN[] = {4, 4};            // Enable Pin, active LOW
byte EN_ACTIVE[] = {LOW, LOW};  // Enable Pin state 
int STEP[] = {6,9};    // Step Pins,  active HIGH
int DIR[] = {7,8};     // Direction Pins, active HIGH
byte DIR_ACTIVE[] = {LOW, LOW};
//int ENCODER_A[] = {0, 0};
//int ENCODER_B[] = {0, 0};

// General initalizers
const int axis = 2;                // Number axis
long axisRes[axis] = {4915200L,4915200L};     // Caculated global resolution for all variables ~~~ Found as lowest common multiple

int stepperRes[axis] = {3200,3200};   // Stepper Resolution
int stepperInc[axis] = {1536,1536};  // Stepper Increments at global LCM resolution

int encoderRes[axis] = {600,600};     // Encoder Resolution
int encoderInc[axis] = {8192,8192};  // Encoder Increments at global LCM resolution

boolean closeEnough[axis] = {false, false};  // Is the actual position within a single step of the setpoint

unsigned int commsRes = 65535U;          // Comms Resolution (16bit) assuming unisigned, not -32,768 to 32,767 (2's Complement);
unsigned int commsInc = 75U;                    // Comms Increments at global LCM resolution

unsigned long calcSetpoint[axis] = {0UL,0UL};      // Caculated setpoint after conversion froms comms data using global LCM resolution
long axisPosition[axis] = {2457600L,2457600L};      // Current position of each axis using global LCM resolution
int axisDirection[axis]  = {0,0};       // Calculated direction of each axis as a +/- stepperInc

// Time keeping
unsigned long currentTime = 0UL;       // Current time
unsigned long lastStep[axis] = {0UL,0UL};         // Last recored step time for each axis

// Comms initializers
const char kAxisName[axis] = { 'P', 'R'};    // the letters ("names") sent from Sim Tools to identify each axis
const char kEOL = '~';                             // End of Line - the delimiter for our acutator values 
const int kMaxCharCount = 5;                       // some insurance to 16 bit...
unsigned int axisSetpoint[axis] = {0U, 0U};             // current Axis setpoint 16bit
int currentAxis = 0;                                   // keep track of the current Axis being read in from serial port
int valueCharCount = 0;                            // how many value characters have we read (must be less than kMaxCharCount!!
// set up some states for our state machine
// psReadAxis = next character from serial port tells us the Axis info that follows
// psReadValue = next characters from serial port until we hit our kMaxCharCount or kEOL which tells us the value
enum TPortState { psReadAxis, psReadValue };   
TPortState currentState = psReadAxis;

// Direction initializers
long halfRes[axis] = {0L,0L};
long limitLowEnd[axis] = {0L,0L};
long limitLowMid[axis] = {0L,0L};
long limitHighMid[axis] = {0L,0L};
long limitHighEnd[axis] = {0L,0L};

// Drive Stepper initializers
int stepTime[axis] = {0,0};
byte halfStep[axis] = {0,0};                  // Half step tracking for safe stepping of motors

// Input / comms
void establishContact() {
  while (Serial.available() <= 0) {
    Serial.println("Waiting...");   // send an initial string
    delay(500);
  }
}  // End establishContact

// Return setpoints as an array
void comms()  {
  char tmpChar;
  int tmpValue;
  // Read serial line for new destination
  while (Serial.available())  {
     // if we're waiting for a Axis name, grab it here
        if (currentState == psReadAxis) {
            tmpChar = Serial.read();
            
#ifdef DEBUG            
Serial.print("read in ");            
Serial.println(tmpChar);            
#endif
        for (int i = 0; i < axis; i++) {    // look for our axis in the array of axis names we set up
        if (tmpChar == kAxisName[i]) {
#ifdef DEBUG            
Serial.print("which is axis ");            
Serial.println(i);            
#endif
                    currentAxis = i;                            // remember which axis we found
                    currentState = psReadValue;                 // start looking for the Axis position 
                    axisSetpoint[currentAxis] = 0;              // initialise the new position
                    valueCharCount = 0;                         // initialise number of value chars read in 
                    break;
                }
            }
        }
        
        // if we're ready to read in the current Axis position data
        if (currentState == psReadValue) {
            while ((valueCharCount < kMaxCharCount) && Serial.available()) {
                tmpValue = Serial.read();
                if (tmpValue != kEOL) {
                    tmpValue = tmpValue - 48;
                    if ((tmpValue < 0) || (tmpValue > 9)) tmpValue = 0;
                    axisSetpoint[currentAxis] = axisSetpoint[currentAxis] * 10 + tmpValue;
                    valueCharCount++;
                }
                else break;
            }
            
            // if we've read the value delimiter, update the Axis and start looking for the next Axis name
            if (tmpValue == kEOL || valueCharCount == kMaxCharCount) {
#ifdef DEBUG            
Serial.print("read in ");            
Serial.println(axisSetpoint[currentAxis]);            
#endif
                
                currentState = psReadAxis;
            }
        }
    }  // End while
}  // End Comms

// Press any key to continue
void pressAnyKey()  {
  Serial.println("Press any key to continue");
  Serial.println("**************************************************************");
  while(Serial.available() == 0)  {
    delay(500);
  }
  delay(50);
  int temp = Serial.read();
  delay(50);
}  // End pressAnyKey

  
// Conversion of input data

  // I need congruity between: 
      // 16bit incoming data
      // Stepper resolution
      // Encoder resolution
      
  //Some useful numbers:
    // 16bit = 0 - 65535
    // 360 degrees at 16bit is 
      // 182.04444444444 bits per degree
      // 0.0054931640625 degrees per bit
      // Lowest common multiple LCM = 14745600 including whole degrees
      // LCM = 4915200 with whole degrees
    
    
    // Encoder is 600ppr
    // Per pulse:
      // 0.6 (degrees)
      // 108 (16bit)
    
    // Stepper drivers is 3200 steps/rev
    // Per step:
      // 0.1125 degrees per step
      // 20.48 bits per step
      
long conversion(int c_axis, unsigned int c_setpoint)  {
  // Declare a setpoint and convert
 unsigned long c_calcSetpoint = 0UL;
  
  // DEBUG
#ifdef DEBUG            
Serial.print("Incoming Setpoint [");
Serial.print(c_axis);
Serial.print("] : ");
Serial.println(c_setpoint);
//Serial.print("Comms Max: ");
//Serial.println(commsRes);
#endif
    
  // Maths:
  // If 16 bit number is not unsigned:
    // Insert maths here!
  
  // Invalid setpoint avoidence
  if (c_setpoint >= 0 && c_setpoint <= commsRes)  {
    // Comms data as 16 bit unsigned
    // Need to convert it to LCM global resolution (32 bit)
    c_calcSetpoint = (unsigned long) c_setpoint * (unsigned long) commsInc;
  }
/*  Very Buggy:  Compiler times out???
  else  {
    c_calcSetpoint = 0;
      
    // DEBUG
#ifdef DEBUG            
Serial.print("Invalid Setpoint [");
Serial.print(c_axis);
Serial.println("] : ");");
#endif

  }
*/  
  
  // DEBUG
#ifdef DEBUG            
Serial.print("Calculated Setpoint [");        
Serial.print(c_axis);
Serial.print("] : ");
Serial.println(c_calcSetpoint);            
#endif

  // Return destination
  return c_calcSetpoint;
  
}  // End conversion
      
// Position detection
void updatePosition()  {
  // Read data from encoder registers
  
  // Compare to internal tracking
  
  // Update internal position if needed
  
}  // End updatePosition

// Direction decision
int decideDirection(int d_axis, long d_setpoint, long d_position)  {
  // Import a setpoint and compare to position
  long d_compare = d_position - d_setpoint;
  
  // Declare directions
  int d_CW = stepperInc[d_axis];
  int d_CCW = -1 * stepperInc[d_axis];
  int d_decision;
  
  // Decide decision
  if (d_compare > limitLowEnd[d_axis] && d_compare < limitLowMid[d_axis])  {
    digitalWrite(DIR[d_axis], !DIR_ACTIVE[d_axis]);
    d_decision = d_CCW;
  }
  else if (d_compare >= limitLowMid[d_axis] && d_compare <= 0)  {
    digitalWrite(DIR[d_axis], DIR_ACTIVE[d_axis]);
    d_decision = d_CW;
  }
  else if (d_compare > 0 && d_compare <= limitHighMid[d_axis])  {
    digitalWrite(DIR[d_axis], !DIR_ACTIVE[d_axis]);
    d_decision = d_CCW;
  }
  else if (d_compare > limitHighMid[d_axis] && d_compare <= limitHighEnd[d_axis])  {
    digitalWrite(DIR[d_axis], DIR_ACTIVE[d_axis]);
    d_decision = d_CW;
  }
  else {
    d_decision = 0;
  }
  
  // Is it within a step?
  if (abs(d_compare) < stepperInc[d_axis])  {
    closeEnough[d_axis] = true;
  }
  else if (abs(d_compare) >= stepperInc[d_axis])  {
    closeEnough[d_axis] = false;
  }
    
  // DEBUG

#ifdef DIRECTIONFEEDBACK            
Serial.print("Direction Decision [");
Serial.print(d_axis);
Serial.print("] : ");            
Serial.println(d_decision);
Serial.print("Comparison [");
Serial.print(d_axis);
Serial.print("] : "); 
Serial.println(d_compare);
#endif


  // Decide on the fastest route (CW or CCW)
  return d_decision;
  
}  // End decideDirection

// PID for speed control
int decideSpeed(int p_axis, long p_setpoint, long p_position)  {
  // Import position and setpoint data
  // Declare variables
  int p_stepTime = 1000;
  
  // PID stuff
  
  // Return step time
  return p_stepTime;
  
  // DEBUG
#ifdef STEPTIMEFEEDBACK
Serial.print("Step Time [");
Serial.print(p_axis);
Serial.print("] : ");            
Serial.println(p_stepTime);
#endif
}  // End decideSpeed

// Drive stepper
void driveStepper(int s_axis, int s_stepTime, long s_position)  {
  // If time elapsed is greater than step time, execute
  // Declare step state
  boolean stepState;
  long s_track = 0L;
  // Import position and step time
  
  // Prepare for step
  if (halfStep[s_axis] == 0)  {
    stepState = HIGH;
    halfStep[s_axis] = 1;
  }
  
  else if (halfStep[s_axis] == 1)  {
    stepState = LOW;
    halfStep[s_axis] = 0;    
  }
  
  // Drive stepper
  digitalWrite(STEP[s_axis], stepState);
  
  // Update internal tracking
  if (halfStep[s_axis] == 0)  {
    // Increment or decrement depending on direction variable
    s_track = s_position + axisDirection[s_axis];
    
    // Constrain to 360 degrees:
    // If greater than resolution
    if (s_track >= axisRes[s_axis])  {
      s_track = s_track - axisRes[s_axis];
    }
    
    // If less than 0
    if (s_track < 0)  {
      s_track = s_track + axisRes[s_axis];
    }
    
    // Update global internal tracked position
    axisPosition[s_axis] = s_track;
  }


  
  // DEBUG
  
#ifdef POSITIONFEEDBACK            
Serial.print("Position [");
Serial.print(s_axis);
Serial.print("] : ");            
Serial.println(s_track);            
#endif

  
  // Update step time...
  
}  // End driveStepper



void setup()  {
  // Comms baud rate
  Serial.begin(115200);
  
  // Define pins
  for (int i=0; i < axis; i++)  {
    pinMode(EN[i], OUTPUT);
    pinMode(STEP[i], OUTPUT);
    pinMode(DIR[i], OUTPUT);
    //pinMode(ENCODER_A[i], INPUT);
    //pinMode(ENCODER_B[i], INPUT);
  }
  
  // Calculate values for decideDirection()
  for (int i=0; i < axis; i++)  {
    halfRes[i] = axisRes[i] / 2;
    limitLowEnd[i] = (-1 * axisRes[i]);
    limitLowMid[i] = (-1 * halfRes[i]);
    limitHighMid[i] = halfRes[i];
    limitHighEnd[i] = axisRes[i];
  }
  
  // Prepare for debug via console
  #ifdef DEBUG
  establishContact();  // send a byte to establish contact until receiver responds
  #endif
  // Until the code needs this to change
  for (int i=0; i < axis; i++)  {
    digitalWrite(EN[i], EN_ACTIVE[i]);  // Enable
  }
  
}  // End setup

void loop()  {
  switch (testType)  {
    case 0:
  
  // Until the code needs this to change
  for (int i=0; i < axis; i++)  {
    digitalWrite(EN[i], EN_ACTIVE[i]);  // Enable
  }
  
  
#ifdef LOOPUPDATE 1
currentTime = micros();
Serial.print("Time: ");
Serial.println(currentTime);

for (int i =0; i < axis; i++)  {
  
  Serial.print("[");
  Serial.print(i);
  Serial.print("] Position: ");
  Serial.println(axisPosition[i]);
  
  Serial.print("[");
  Serial.print(i);
  Serial.print("] Setpoint: ");
  Serial.println(calcSetpoint[i]);
}
#endif
  
  // First establish setpoints for each axis from serial data
  if (Serial.available() > 0)  {
    // Read data
    comms();
    
    // Convert data
    for (int i = 0; i < axis; i++)  {
      calcSetpoint[i] = conversion(i, axisSetpoint[i]);
    }
    // Setpoints are now calcSetpoint[axis]
    
  }  // End if(Serial.available() > 0)
  
  
    // Decide on shortest path direction for each axis
    for (int i = 0; i < axis; i++)  {
      axisDirection[i] = decideDirection(i, calcSetpoint[i], axisPosition[i]);  //1536;
    }

  
  // Establish an acceleration profile for each axis
  
  for (int i = 0; i < axis; i++)  {
    stepTime[i] = decideSpeed(i, calcSetpoint[i], axisPosition[i]);
  }
  
  
  
  // If ready, drive the stepper motors
  for (int i = 0; i < axis; i++)  {
    
    // If the axis position is not at the setpoint    
    if (axisPosition[i] != calcSetpoint[i] && closeEnough[i] == false)  {
         
      // Update time
      currentTime = micros();
    
      // If the correct time has passed:
      if (currentTime - lastStep[i] > stepTime[i])  {
      
        // Update this step time
        lastStep[i] = currentTime;
      
        // Take a step
        driveStepper(i, stepTime[i], axisPosition[i]);
      }
    }
  }

#ifdef USERCONTROL
pressAnyKey();  // wait for user input
#endif
  break;  // Case 0
  
  case 1: // Speed Control
  
  // Enable Pins
  
  // Check for input
    
    // Update step time
    
  // Drive stepper if step_time > 0 
  
  break;  // Case 1
  }
}  // End loop

